// Copyright (C) 2016 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only

#include <QtCore/QCoreApplication>
#include <QtQuickTest/quicktest.h>

QUICK_TEST_MAIN(tst_qtexttospeech_qml)
